/* Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * dsn.c
 *
 *  Created on: Feb 21, 2010
 *      Author: drewyrr
 */

#include <apr_lib.h>

#include "dsn.h"
#include "dsndbd.h"

apr_status_t dsn_process_message(mimeparse_rec *conf, mime_rec *mime, int last) {
	const char *content_type;
	dsn_rec *dsn;
	notification_rec *notification;
	void *dummy = NULL;
	apr_status_t status = APR_SUCCESS;

	apr_pool_userdata_get(&dummy, "dsn", conf->pool);
	dsn = dummy;
	notification = dsn->notification;

	if (!notification) {
		dsn->notification = notification = notification_make(conf->pool, conf);
	}

	content_type = apr_table_get(mime->headers, "Content-Type");
	if (content_type && dsn) {

		/* the delivery report */
		if (!strcmp(content_type, "message/delivery-status")) {

			apr_status_t status;

			/* Reporting-MTA: dns; mta.example.com
			 * X-Postfix-Queue-ID: BAEFA79BA1
			 * X-Postfix-Sender: rfc822; mail-bounces+dontexist=example.com@example.com
			 * Arrival-Date: Sun,  7 Feb 2010 18:24:08 -0600 (CST)
			 *
			 * Final-Recipient: rfc822; dontexist@example.com
			 * Original-Recipient: rfc822;dontexist@example.com
			 * Action: failed
			 * Status: 5.1.1
			 * Diagnostic-Code: x-unix; Invalid user specified.
			 */
			status = mime_parse_headers(conf, mime->body,
					notification->delivery_status);
			if (APR_SUCCESS == status || APR_EOF == status) {
				status = mime_parse_headers(conf, mime->body,
						notification->delivery_status_body);
				if (APR_SUCCESS == status || APR_EOF == status) {
					status = APR_SUCCESS;
				}
			}

		}

		/* original message headers */
		else if (!strcmp(content_type, "message/rfc822") || !strcmp(content_type, "text/rfc822-headers")) {

			/* Received: from mta.example.com (localhost [127.0.0.1])
			 *         by mta.example.com (Postfix) with ESMTP id BAEFA79BA1
			 *         for <dontexist@example.com>; Sun,  7 Feb 2010 18:24:08 -0600 (CST)
			 * Received: by mta.example.com (Postfix, from userid 500)
			 *         id AB66D79DB3; Sun,  7 Feb 2010 18:24:08 -0600 (CST)
			 * To: dontexist@example.com
			 * From: drewyrr@gmail.com
			 * Subject: test1
			 * Message-Id: <20100208002408.AB66D79DB3@mta.example.com>
			 * Date: Sun,  7 Feb 2010 18:23:48 -0600 (CST)
			 * X-Virus-Scanned: ClamAV using ClamSMTP
			 *
			 */
			char *body;
			apr_size_t bodylen;
			apr_brigade_pflatten(mime->body, &body, &bodylen,
					notification->pool);

			status = mime_parse_headers(conf, mime->body, notification->rfc822);
			if (APR_SUCCESS == status || APR_EOF == status) {
				status = APR_SUCCESS;
			}

		}

		/* plain text message */
		else if (!strcmp(content_type, "text/plain") || !strncmp(content_type,
				"text/plain;", 11)) {
			char *body;
			apr_size_t bodylen;

			/* This is the mail system at host mta.example.com.
			 *
			 * I'm sorry to have to inform you that your message could not
			 * be delivered to one or more recipients. It's attached below.
			 *
			 * For further assistance, please send mail to <postmaster>
			 *
			 * If you do so, please include this problem report. You can
			 * delete your own text from the attached returned message.
			 *
			 *                    The mail system
			 *
			 * <dontexist@example.com>: user unknown. Command output: Invalid user specified.
			 */
			apr_brigade_pflatten(mime->body, &body, &bodylen,
					notification->pool);
			apr_brigade_write(notification->body, NULL, NULL, body, bodylen);

		}

		/*
		 * printf("%d/%s: %s\n", mime->level, last == MESSAGE_NORMAL ? "normal"
		 * : last == MESSAGE_LAST ? "last" : "none", content_type);
		 */
	}

	/* commit what we have found so far */
	if (!mime->level && MESSAGE_LAST == last) {
		const char *to, *recipient, *message_id;
		apr_uint64_t recipient_key = 0;
		notification->headers = mime->headers;

		/* get the sender */
		to = apr_table_get(mime->headers, "To");
		recipient = verp_parse(notification->pool, to, dsn->verp);

		/* get the notification message-id */
		message_id = apr_table_get(mime->headers, "Message-ID");

		/* verify the hash we inserted into the original message (if configured) */

		/* store the message... */
		status = dsndbd_update_message(notification->pool, dsn->dsncon,
				recipient, notification->rfc822, &recipient_key);
		if (APR_SUCCESS == status) {

			/* ...and the notification */
			status = dsndbd_update_notification(notification->pool,
					dsn->dsncon, message_id, recipient_key, notification->delivery_status,
					notification->delivery_status_body, notification->body);

		}

		/* we're done, reset the processdsn for the next message (if any) */
		apr_pool_destroy(notification->pool);
		dsn->notification = NULL;
	}

	return status;
}

dsn_rec *dsn_make(apr_pool_t *parent) {
	apr_pool_t *pool;
	dsn_rec *dsn;

	apr_pool_create(&pool, parent);
	dsn = apr_pcalloc(pool, sizeof(dsn_rec));
	dsn->pool = pool;

	dsn->mimeparse = mime_parse_make(pool);
	dsn->mimeparse->process_message = &dsn_process_message;

	apr_pool_userdata_setn(dsn, "dsn", NULL, dsn->mimeparse->pool);

	return dsn;
}

notification_rec *notification_make(apr_pool_t *parent,
		mimeparse_rec *mimeparse) {
	apr_pool_t *pool;
	notification_rec *notification;

	apr_pool_create(&pool, parent);
	notification = apr_pcalloc(pool, sizeof(notification_rec));
	notification->pool = pool;

	notification->delivery_status = apr_table_make(notification->pool, 10);
	notification->delivery_status_body = apr_table_make(notification->pool, 10);
	notification->rfc822 = apr_table_make(notification->pool, 10);
	notification->body = apr_brigade_create(notification->pool,
			mimeparse->alloc);

	return notification;
}
