/* Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * mimeparse.c
 *
 *  Created on: Feb 10, 2010
 *      Author: drewyrr
 */

#include "mimeparse.h"
#include <apr_strings.h>
#include <apr_lib.h>

apr_status_t mime_parse_header_value(mimeparse_rec *conf, const char *raw,
		char **value, apr_table_t *params) {
	char *buf = apr_pstrdup(conf->pool, raw);
	*value = buf;
	buf = strchr(buf, ';');
	if (buf) {
		*(buf++) = 0;
		while (buf) {
			char *val, *key;
			for (; *buf == ' '; buf++)
				;
			if (!*buf) {
				break;
			}
			key = buf;
			val = strchr(buf, '=');
			if (val) {
				int match = 0, len;
				*(val++) = 0;
				if (*val == '\"') {
					val++;
					match = 1;
				}
				buf = strchr(val, ';');
				if (buf) {
					*(buf++) = 0;
				}
				len = strlen(val);
				if (*(val + len - 1) == '\"') {
					if (match) {
						*(val + len - 1) = 0;
					} else {
						return APR_EGENERAL;
					}
				}
				if (params) {
					apr_table_addn(params, key, val);
				}
			} else {
				return APR_EGENERAL;
			}
		}
	}
	return APR_SUCCESS;
}

// FIXME: look at the pool we are allocated from
apr_status_t mime_parse_header(mimeparse_rec *conf, apr_bucket_brigade *bb,
		apr_table_t *headers) {
	char *key;
	apr_off_t len;
	apr_size_t size;
	char *value;

	apr_brigade_length(bb, 1, &len);
	key = apr_palloc(conf->pool, len + 1);
	key[len] = 0;
	size = len;
	apr_brigade_flatten(bb, key, &size);

	value = strchr(key, ':');
	if (value) {
		*(value++) = 0;
		for (; *value == ' '; value++)
			;
		apr_table_addn(headers, key, value);
		apr_brigade_cleanup(bb);
		return APR_SUCCESS;
	} else {
		/* header expected, but none found */
		return APR_EGENERAL;
	}

}

// FIXME: look at the pool we are allocated from
apr_status_t mime_parse_headers(mimeparse_rec *conf, apr_bucket_brigade *bb,
		apr_table_t *headers) {
	apr_bucket_brigade *obb, *line;
	apr_status_t status;
	char str[HUGE_STRING_LEN];

	obb = apr_brigade_create(conf->pool, conf->alloc);
	line = apr_brigade_create(conf->pool, conf->alloc);

	do {
		apr_size_t len = HUGE_STRING_LEN;
		apr_off_t off;

		/* deal with a previous line */
		apr_brigade_length(line, 1, &off);

		/* this will always succeed */
		status = apr_brigade_split_line(obb, bb, APR_BLOCK_READ,
				HUGE_STRING_LEN);
		if (APR_SUCCESS != status) {
			return status;
		}

		apr_brigade_flatten(obb, str, &len);

		/* end of buffer, leave */
		if (!len) {
			if (off) {
				status = mime_parse_header(conf, line, headers);
			}
			return APR_EOF;
		}

		/* cut off the CRLF */
		while (len && (str[len - 1] == 13 || str[len - 1] == 10)) {
			len--;
		}

		/* blank line, we're done */
		if (!len) {
			if (off) {
				status = mime_parse_header(conf, line, headers);
			}
			return APR_SUCCESS;
		}

		/* continuation line, add to buffer */
		else if ((str[0] == ' ' || str[0] == '\t')) {
			apr_brigade_write(line, NULL, NULL, str + 1, len - 1);
		}

		/* normal line */
		else {
			if (off) {
				status = mime_parse_header(conf, line, headers);
			}
			apr_brigade_write(line, NULL, NULL, str, len);
		}

		apr_brigade_cleanup(obb);

	} while (1);

}

mimeparse_rec *mime_parse_make(apr_pool_t *parent) {
	apr_pool_t *pool;
	mimeparse_rec *mimeparse;

	apr_pool_create(&pool, parent);
	mimeparse = apr_pcalloc(pool, sizeof(mimeparse_rec));
	mimeparse->pool = pool;

	apr_file_open_stderr(&mimeparse->err, pool);
	apr_file_open_stdin(&mimeparse->in, pool);
	mimeparse->alloc = apr_bucket_alloc_create(pool);

	return mimeparse;
}

mime_rec *mime_make(apr_pool_t *ppool, mimeparse_rec *conf, mime_rec *parent) {
	apr_pool_t *pool;
	mime_rec *mime;

	apr_pool_create(&pool, ppool);
	mime = apr_pcalloc(pool, sizeof(mime_rec));
	mime->pool = pool;

	mime->phase = FIND_START;
	if (parent) {
		mime->level = parent->level + 1;
	}
	mime->line = apr_brigade_create(conf->pool, conf->alloc);
	mime->headers = apr_table_make(conf->pool, 10);
	mime->body = apr_brigade_create(conf->pool, conf->alloc);

	return mime;
}

void mime_reset(mimeparse_rec *conf, mime_rec *mime) {
	mime_rec *child;

	mime->phase = FIND_FROM;
	apr_brigade_cleanup(mime->line);
	apr_table_clear(mime->headers);
	apr_brigade_cleanup(mime->body);

	child = mime->first;
	while (child) {
		mime_rec *next = child->next;
		apr_pool_destroy(child->pool);
		child = next;
	}

	mime->first = NULL;
	mime->last = NULL;

}

apr_status_t mime_last_line(mimeparse_rec *conf, mime_rec *mime) {
	apr_status_t status = APR_SUCCESS;

	if (conf->process_message) {
		status = conf->process_message(conf, mime, MESSAGE_LAST);
	}

	return status;
}

apr_status_t mime_process_line(mimeparse_rec *conf, mime_rec *mime, const char *str,
		apr_size_t len) {
	apr_status_t status = APR_SUCCESS;
	apr_off_t off;

	switch (mime->phase) {
	case FIND_START:
	case FIND_FROM: {
		if (len > 5 && !strncmp(str, "From ", 5)) {

			if (FIND_FROM == mime->phase && conf->process_message) {
				status = conf->process_message(conf, mime, MESSAGE_LAST);
			}

			mime_reset(conf, mime);
			mime->phase = FIND_HEADERS;

		} else {

			apr_brigade_write(mime->body, NULL, NULL, str, len);

			if (conf->process_line) {
				status = conf->process_line(conf, mime, str, len);
			}

		}
		break;
	}
	case FIND_HEADERS: {

		/* cut off the CRLF */
		while (len && (str[len - 1] == 13 || str[len - 1] == 10)) {
			len--;
		}

		/* blank line, switch to find from */
		if (!len) {
			const char *content_type;

			/* deal with a previous line */
			apr_brigade_length(mime->line, 1, &off);
			if (off) {
				status = mime_parse_header(conf, mime->line, mime->headers);
			}

			mime->phase = FIND_FROM;

			/* deal with multipart/ * */
			content_type = apr_table_get(mime->headers, "Content-Type");
			if (content_type) {
				char *value;
				apr_table_t *params = apr_table_make(conf->pool, 10);
				status = mime_parse_header_value(conf, content_type, &value, params);
				if (APR_SUCCESS == status && !strncmp(content_type,
						"multipart/", 10)) {
					mime->boundary = apr_table_get(params, "boundary");
					if (mime->boundary) {
						mime->boundary_length = strlen(mime->boundary);
						mime->phase = FIND_BOUNDARY;
					}
				}
			}

		}

		/* continuation line, add to buffer */
		else if (len && (str[0] == ' ' || str[0] == '\t')) {
			apr_brigade_write(mime->line, NULL, NULL, str + 1, len - 1);
		}

		/* normal line */
		else {
			/* deal with a previous line */
			apr_brigade_length(mime->line, 1, &off);
			if (off) {
				status = mime_parse_header(conf, mime->line, mime->headers);
			}
			apr_brigade_write(mime->line, NULL, NULL, str, len);
		}
		break;
	}
	case FIND_BOUNDARY: {

		/* boundary line? */
		if (len >= (mime->boundary_length + 2) && str[0] == '-' && str[1]
				== '-' && !strncmp(str + 2, mime->boundary,
				mime->boundary_length)) {
			/* terminating boundary line? */
			if (len >= (mime->boundary_length + 4) && str[mime->boundary_length
					+ 2] == '-' && str[mime->boundary_length + 3] == '-') {

				/* process the last message, if present */
				if (mime->last) {
					if (conf->process_message) {
						status = conf->process_message(conf, mime->last,
								MESSAGE_LAST);
					}
				}

				mime->phase = FIND_FROM;

			}
			/**/
			else {
				mime_rec *next = mime_make(mime->pool, conf, mime);
				next->phase = FIND_HEADERS;

				/* process the last message, if present */
				if (mime->last) {
					if (conf->process_message) {
						status = conf->process_message(conf, mime->last,
								MESSAGE_NORMAL);
					}
					mime->last->next = next;
				}

				/* set up a new mime child */
				if (!mime->first) {
					mime->first = next;
				}
				next->parent = mime;
				mime->last = next;
			}
		}

		/* otherwise nested */
		else {
			if (mime->last) {
				status = mime_process_line(conf, mime->last, str, len);
			}
		}

	}
	}
	return status;
}

/**
 * Convert mail-bounces+dontexist=example.com@example.com to dontexist@example.com
 */
const char *verp_parse(apr_pool_t *pool, const char *verp, const char *sep) {
	const char *user, *domain, *end;
	char *address;

	if (!verp) {
		return NULL;
	}

	if (!sep) {
		sep = "+=";
	}

	user = strchr(verp, sep[0]);
	if (user) {
		user++;
		domain = strchr(user, sep[1]);
		if (domain) {
			domain++;
			end = strchr(domain, '@');
			if (end) {
				int offset = 0;

				address = apr_palloc(pool, end - user + 1);

				memcpy(address, user, domain - user - 1);
				offset += domain - user - 1;
				address[offset++] = '@';
				memcpy(address + offset, domain, end - domain);
				offset += end - domain;
				address[offset] = 0;

				return address;
			}
		}
	}

	return NULL;

}
